# Copyright (c) 2008-2009 Code Lounge.
# See LICENSE for details

import time, types, inspect
from textwrap import dedent

from zope.interface import implements

from twisted.python import log
from twisted.python.components import registerAdapter
from twisted.internet import reactor

from gina.igina import ICommand
from gina.errors import UsageError

from gina.util import usage, HelpMixin

import config

class CommandLookupMixin(object):
    implements(ICommand)

    name = None
    usage = 'No usage information'
    help = """No additional help."""

    def locateCommand(self, params):
        """
        Find the handler responsible for a given command.

        The first parameter is popped from C{params} and is treated as the
        command handler name. Handlers should be attributes with the form
        C{cmd_COMMAND}.  If no handler is found for C{cmd} then L{UsageError}
        is raised.

        @param params: The parameters, including the command, for a command
        @type params: C{list}

        @return: The command handler for C{cmd} and the remaining parameters
        @rtype: C{(ICommand, list)}
        """
        full_cmd = ' '.join(params)
        cmd = params.pop(0)
        method = getattr(self, 'cmd_%s' % cmd.lower(), None)
        if method is not None:
            return ICommand(method), params
        else:
            if len(params) > 0:
                subcommand = params.pop(0)
                subcommand = getattr(self, 'cmd_%s_%s' % (cmd.lower(),
                    subcommand.lower()), None)
                if subcommand is not None:
                    return ICommand(subcommand), params

        #msg = """I couldn't find a command or feed alias matching '%s'.
        #Try '<b>help</b>' for help or '<b>help [command]</b>' for
        #additional help on a specific command."""
        #msg = ' '.join([l.strip() for l in msg.splitlines()])
        #raise UsageError(msg % (full_cmd,))


class MethodCommand(object):
    implements(ICommand)

    defaultUsage = 'No usage information'
    defaultHelp = """No additional help."""

    def __init__(self, method):
        super(MethodCommand, self).__init__()

        usage = getattr(method, 'usage', None)
        if usage is None:
            usage = self.defaultUsage

        help = getattr(method, 'help', None)
        if help is None:
            help = self.defaultHelp

        lines = [dedent(ln) for ln in help.splitlines()]
        if not lines[0]:
            lines.pop(0)

        self.method = method
        self.params = []
        self.name = method.__name__[len('cmd_'):]
        self.usage = usage
        self.shortHelp = lines[0]
        self.help = '\n'.join(lines)

        self.minargs, self.maxargs = self.getArgLimits()

    def getArgLimits(self):
        minargs, maxargs = arglimits = getattr(self.method, 'arglimits',
                (None, None))

        args, vararg, varkw, defaults = inspect.getargspec(self.method)
        # Exclude self and source parameters.
        normArgCount = len(args) - 2

        if minargs is None:
            minargs = normArgCount

        if maxargs is None:
            if vararg is None:
                maxargs = normArgCount
            else:
                maxargs = None

        return minargs, maxargs

    def locateCommand(self, params):
        self.params = params
        return self, []

    def invoke(self, source):
        numargs = len(self.params)

        if numargs < self.minargs:
            raise UsageError('Not enough arguments' % self.params)
        if self.maxargs is not None and numargs > self.maxargs:
            raise UsageError('Too many arguments')

        return self.method(source, *self.params)

registerAdapter(MethodCommand, types.MethodType, ICommand)


class EventHandler(object):
    """
    An event dispatch mechanism.
    """
    def locateHandler(self, event):
        """
        Locate a handler for the specified method.

        Handlers should be named according to the pattern C{event_EVENTNAME}
        and arguments are passed as positional parameters.  Any errors raised
        by event handlers are swallowed and logged.  If no handler is present,
        the event is ignored.

        An example event handler::

            def event_message(self, source, dest, msg): ...

        @returns: a callable handler, or C{None}
        """
        return getattr(self, 'event_%s' % (event,), None)


class PseudoHandler(EventHandler, CommandLookupMixin, HelpMixin):
    help = ''

    def setController(self, controller):
        self.ctrl = controller
        self.store = self.ctrl.store
        self.protocol = self.ctrl.protocol
        if hasattr(self, 'postInit'):
           self.postInit()
        return self

    def event_message(self, source, message):
        self.command(source, message)

    def command(self, source, msg):
        """
        Dispatch a command.
        """
        params = msg.split()
        try:
            cmd, params = self.locateCommand(params)
   
            try:
                # Save params to the adapted ICommand() cmd and invoke
                cmd, params = cmd.locateCommand(params)
                return cmd.invoke(source)
 
            except UsageError, e:
                self.protocol.msg(source, 'Invalid command usage. %s' % str(e))
                if getattr(cmd, 'usage', None):
                    self.protocol.msg(source, 'Usage: %s' % (cmd.usage,))
        
        except UsageError, e:
            self.protocol.msg(source, str(e))
       
        except TypeError:
            pass

        except:
            log.err()


class GinaController(EventHandler):
    def __init__(self, store, protocol, params):
        self.store = store
        self.params = params
        self.handlers = [self]
        self.protocol = protocol
        self.users = {}
        
    def addHandler(self, service):
        self.handlers.append(service)
        
    def listHandlers(self):
        return self.handlers
    
    def dropHandlers(self):
        # XXX
        # Handlers should clean up after themselves (remove/callLater(remove) pseudoclients 
        # Not sure where to do this though...
        self.handlers = [self]

    def event(self, event, *params):
        for p in self.handlers:
            self.sendEvent(p, event, *params)

    def sendEvent(self, receiver, event, *params):
        """
        Dispatch an event to a handler, if it exists.
        """
        method = receiver.locateHandler(event)
        if method is None:
            return

        try:
            method(*params)
        except:
            log.err()

    def getUser(self, nick):
        return self.users.get(nick.lower())

    def createUser(self, user):
        nick = user['nick']
        u = self.users[nick.lower()] = User(
            nick=nick,
            user=user['username'],
            host=user['hostname'],
            name=user['realname'],
            modes=set(user['modes']),
            signon=user['timestamp'],
            server=user['server'])

        self.event('newuser', u)
       
    def event_stats(self, source):
        self.protocol.msg(source, 'stats go here')

    def event_connected(self, protocol):
        pass
    
    def event_quit(self, nick, msg):
        user = self.getUser(nick)
        self.users.pop(nick.lower())

