# -*- coding: utf-8 -*-
# Copyright (c) 2008-2009 Code Lounge.
# See LICENSE for details

import time

from twisted.python import log
from twisted.internet.protocol import ReconnectingClientFactory
from twisted.protocols.basic import LineOnlyReceiver
from twisted.words.protocols.jabber import client, jid, xmlstream
from twisted.words.xish import domish

from twisted.internet import reactor, threads
from twisted.internet.task import LoopingCall

import config

class JabberBadMessage(ValueError):
    """
    Raised when a malformed or invalid XMPP message is encountered.
    """

def parsemsg(s):
    """
    Parses a raw Jabber message from the user and returns the components.

    Empty messages are treated as errors and raise L{JabberBadMessage}.

    @param s: Raw XMPP message
    @type s: C{str}

    @return: The command and additional arguments
    @rtype: C{(str, list)}
    """
    if not s:
        raise JabberBadMessage('Empty line.')

    args = s.split()
    command = args.pop(0)
    
    if command == 'empty':
        raise JabberBadMessage('Empty line.')

    return command, args


class JabberProtocol(xmlstream.XmlStream):
    def __init__(self, jid, secret):
        xmlstream.XmlStream.__init__(self,
                client.BasicAuthenticator(jid, secret))
        self.me = jid
        self.addObserver(xmlstream.STREAM_CONNECTED_EVENT, self.connected)
        self.addObserver(xmlstream.STREAM_AUTHD_EVENT, self.authd)
        
        AUTH_FAILED_EVENT = "//event/client/basicauth/authfailed"
        self.addObserver(AUTH_FAILED_EVENT, self.authFailedEvent)

    def connected(self, xmlstream):
        self.xmlstream = xmlstream
        self.params = self.controller.params
        self.startHandshake()
        self.controller.event('connected', self)

    def authd(self, xmlstream):
        print "Authentication successful."

        # Send presence so clients know we're actually online
        print 'Sending online presence.'
        presence = domish.Element(('jabber:client', 'presence'))
        presence.addElement('status').addContent('Online')
        xmlstream.send(presence)
        
        # Run initialization to set up event observers
        self.initOnline()

    def authFailedEvent(self, xmlstream):
        print 'Auth failed!'
        reactor.stop()

    def initOnline(self):
        print 'Add event observers...'
        self.xmlstream.addObserver('/message', self.gotMessage)
        self.xmlstream.addObserver('/presence', self.gotPresence)
        self.xmlstream.addObserver('/*', self.gotSomething)
        
        # Launch feed scheduler
        #rss_reader() # bootstrap
        
        print 'Ready to handle requests.'

    def startHandshake(self):
        pass

    def gotMessage(self, el):
        if config.DEBUG:
            print el.toXml()

        # Strip off Jabber resource to we can identify the user no matter 
        # where they login from
        if '/' in el["from"]:
            el["from"] = el["from"].split('/')[0]
    
        # Grab the body of the message
        body = 'empty'
        for e in el.elements():
            if e.name == "body":
                body = e.__str__()
                break
        orig = body
        
        # Command lookup
        try:
            command, params = parsemsg(body)
        except JabberBadMessage, e:
            # This happens when a user types and the 'typing' chat state is 
            # sent.
            pass
        else:
            self.handleCommand(el['from'], 'message', [body])


    def handleCommand(self, source, command, params):
        method = getattr(self, 'xmpp_%s' % (command.upper(),), None)
        try:
            def handle():
                if method is not None:
                    method(source, *params)
                else:
                    self.xmpp_unknown(source, command, params)
            self.controller.store.transact(handle)
        except JabberBadMessage, e:
            raise
        except:
            log.msg('%s|%s - %s' % (source, command, params))
            log.err()

    def gotPresence(self, el):
        print "gotPresence(el=%s)" % str(el.getAttribute('type'))
        if el.getAttribute('type') == 'subscribe':
            self.sendMsg('presence', el.attributes['from'])
            self.controller.event('new_user', el.attributes['from'])

    def gotSomething(self, el):
        if config.DEBUG:
            print "<<<", el.toXml()

    def sendMsg(self, type, to, msg=None):
        message = domish.Element(('jabber:client',type))
        message["from"] = self.me.full()
        message["to"] = jid.JID(to).full()

        if type == "message":
            message["type"] = "chat"
            message.addElement("body", "jabber:client", msg)
        elif type == "presence":
            message["type"] = "subscribed"

        """Create Message"""
        msg = message.toXml() #.addRawXml("<body>" + msg + "</body>")

        """Add Bold Text"""
        msg = msg.replace("&lt;b&gt;", "<span style='font-weight: bold;'>")
        msg = msg.replace("&lt;/b&gt;", "</span>")

        """Shorten links"""
        msg = msg.replace("&lt;gina-link&gt;", "<a href='")
        msg = msg.replace("&lt;/gina-link&gt;", "'>More...</a>")

        """Red text"""
        msg = msg.replace("&lt;gina-red&gt;", "<span style='color: #FF0000;'>")
        msg = msg.replace("&lt;/gina-red&gt;", "</span>")

        """Orange text"""
        msg = msg.replace("&lt;gina-orange&gt;", "<span style='color: #FF8104;'>")
        msg = msg.replace("&lt;/gina-orange&gt;", "</span>")

        """Green text"""
        msg = msg.replace("&lt;gina-green&gt;", "<span style='color: #009D00;'>")
        msg = msg.replace("&lt;/gina-green&gt;", "</span>")

        # Drop trailing newlines
        msg = msg.strip()

        if config.DEBUG:
            print '>>> ' + msg

        self.xmlstream.send(msg)

    def msg(self, to, msg):
        self.sendMsg('message', to, msg)

    ### Handlers

    def xmpp_unknown(self, source, command, params):
        if config.DEBUG:
            log.msg('UNKNOWN COMMAND: %s - %s' % (command, params))

    def xmpp_CONNECT(self, user):
        pass

    def xmpp_DISCONNECT(self, user):
        pass

    def xmpp_MESSAGE(self, source, message):
        self.controller.event('message', source, message)


class JabberFactory(ReconnectingClientFactory):
    """
    A simple client factory for gina.

    We inherit from ReconnectingClientFactory to ensure Gina tries to stay 
    connected at all times.
    """
    maxDelay = 5 * 60

    def __init__(self, controllerFactory, protocolFactory, params):
        self.controllerFactory = controllerFactory  # GinaController
        self.protocolFactory   = protocolFactory    # JabberProtocol
        self.params = params                        # Config data

    def buildProtocol(self, addr):
        """
        Create a protocol instance.

        We give the protocol a controller attribute, because the factory doesn't
        have any functionality the protocol needs to care about.
        """
        # Reset reconnect counter - This probably shouldn't be here - connectionMade() would be better
        self.resetDelay()

        # Setup protocol
        jabberid = jid.JID('%s/Gina' % self.params['JABBERID'])
        p = self.protocolFactory(jabberid, self.params['PASSWORD'])
        p.factory = self
        p.controller = self.controllerFactory(p)
        return p
