# Copyright (c) 2008-2009 Code Lounge.
# See LICENSE for details

from gina.igina import ICommand
from gina.errors import UsageError

from twisted.internet import reactor, task, error as ineterror
from twisted.web import client, http, error as weberror
from twisted.python import log

from nevow.url import URL
from nevow.rend import Page, Fragment


def usage(desc, min=None, max=None):
    """
    Decorate a function with usage information.

    After decorating a function with C{usage} it gains 3 new attributes::

        C{usage}::
            A usage string generated from C{desc}.

        C{help}::
            A help string copied and dedented from C{__doc__}.

        C{arglimits}::
            A tuple of the minimum and maximum arguments from C{min} and
            C{max}, or C{None}.

    @param desc: A description of the usage. Angle brackets C{(<>)} should be
                 used to denote mandatory arguments while square brackets C{([])}
                 should be used for optional arguments.
    @type desc: C{str}

    @param min: The minimum number of arguments to enforce or C{None}. C{None}
                means that the implementation should decide on a suitable
                minumum.
    @type min: C{int} or C{None}

    @param max: The maximum number of arguments to enforce or C{None}. C{None}
                means that the implementation should decide on a suitable
                maximum.
    @type min: C{int} or C{None}
    """
    def fact(f):
        f.usage = desc
        f.help = f.__doc__
        f.arglimits = (min, max)
        return f
    return fact


def formatHelpOverview(commands):
    if len(commands) == 0:
        return

    mnl = max(len(c.name) for c in commands)
    for cmd in commands:
        name = cmd.name
        padding = ' ' * (mnl - len(name))
        if hasattr(cmd, 'shortHelp') is False:
            cmd.shortHelp = cmd.help.splitlines()[0]
                
        yield '%s  %s%s' % (name.upper().replace('_', ' '), padding,
                cmd.shortHelp)


class HelpMixin(object):
    _hiddenHelpOverviews = ['cmd_help', 'cmd_die']
    help = ''

    @property
    def commands(self):
        commands = [ICommand(getattr(self, attrName)) for attrName in dir(self)
                    if attrName.startswith('cmd_')
                    and attrName not in self._hiddenHelpOverviews]

        commands.sort(key=lambda x: x.name)
        return commands

    def showHelpOverview(self, source, params, cmdset=None):
        if cmdset is None:
            cmdset = self
 
        if params:
            try:
                #print params
                #cmd, junk = cmdset.locateCommand([params[0]])
                cmd, junk = cmdset.locateCommand(list(params))
                #print repr(cmd), repr(junk)
                self.protocol.msg(source, 'Usage: %s' % (cmd.usage,))
                self.protocol.msg(source, cmd.help)
            except UsageError, e:
                self.protocol.msg(source, 'Invalid command usage. %s' % str(e))
        else:
            response = ''
            for r in formatHelpOverview(cmdset.commands):
                response += '%s\n' % r
            self.protocol.msg(source, response)

    @usage('HELP [command]')
    def cmd_help(self, source, *params):
        if len(params) == 0:
            # Standalone help
            self.protocol.msg(source, self.help)
        
        self.showHelpOverview(source, params)

class PerseverantDownloader(object):
    """
    Perseverantly attempt to download a URL.

    Each retry attempt is delayed by L{factor} up to a maximum of L{maxDelay},
    starting at L{initialDelay}.

    @type url: C{nevow.url.URL}
    @ivar url: The HTTP URL to attempt to download

    @type maxDelay: C{float}
    @cvar maxDelay: Maximum delay, in seconds, between retry attempts

    @type initialDelay: C{float}
    @cvar initialDelay: The delay before the first retry attempt

    @type factor: C{float}
    @cvar factor: The factor to increase the delay by after each attempt

    @type retryableHTTPCodes: C{list}
    @cvar retryableHTTPCodes: HTTP error codes that suggest the error is
        intermittent and that a retry should be attempted

    @type defaultTimeout: C{float}
    @cvar defaultTimeout: Default fetch timeout value
    """
    maxDelay = 3600
    initialDelay = 1.0
    factor = 1.6180339887498948

    retryableHTTPCodes = [408, 500, 502, 503, 504]

    defaultTimeout = 300.0

    def __init__(self, url, tries=10, timeout=defaultTimeout, *a, **kw):
        """
        Prepare the download information.

        Any additional positional or keyword arguments are passed on to
        C{twisted.web.client.HTTPPageGetter}.

        @type url: C{nevow.url.URL} or C{unicode} or C{str}
        @param url: The HTTP URL to attempt to download

        @type tries: C{int}
        @param tries: The maximum number of retry attempts before giving up

        @type timeout: C{float}
        @param timeout: Timeout value, in seconds, for the page fetch;
            defaults to L{defaultTimeout}
        """
        if isinstance(url, unicode):
            url = url.encode('utf-8')
        if isinstance(url, str):
            url = URL.fromString(url)

        self.url = url.anchor(None)
        self.args = a
        self.kwargs = kw
        self.delay = self.initialDelay
        self.tries = tries
        self.timeout = timeout

    def __repr__(self):
        return '<%s %s>' % (type(self).__name__, self.url)

    def go(self):
        """
        Attempt to download L{self.url}.
        """
        d, f = getPage(str(self.url), timeout=self.timeout, *self.args, **self.kwargs)
        return d.addErrback(self.retryWeb
               ).addCallback(lambda data: (data, f.response_headers))

    def retryWeb(self, f):
        """
        Retry failed downloads in the case of "web errors."

        Only errors that are web related are considered for a retry attempt
        and then only when the HTTP status code is one of those in
        L{self.retryableHTTPCodes}.

        Other errors are not trapped.
        """
        f.trap((weberror.Error, ineterror.ConnectionDone))
        err = f.value
        if int(err.status) in self.retryableHTTPCodes:
            return self.retry(f)

        return f

    def retry(self, f):
        """
        The retry machinery.

        If C{self.tries} is greater than zero, a retry is attempted for
        C{self.delay} seconds in the future.
        """
        self.tries -= 1
        log.msg('PerseverantDownloader is retrying, %d attempts left.' % (self.tries,))
        log.err(f)
        self.delay = min(self.delay * self.factor, self.maxDelay)
        if self.tries == 0:
            return f

        return task.deferLater(reactor, self.delay, self.go)


def handle206(f):
    """
    Return any partial content when HTTP 206 is returned.
    """
    f.trap(weberror.Error)
    err = f.value
    try:
        if int(err.status) == http.PARTIAL_CONTENT:
            return err.response
    except ValueError:
        pass

    return f


def getPage(url, contextFactory=None, *args, **kwargs):
    if 'agent' not in kwargs:
        kwargs['agent'] = 'Gina Page Fetcher'

    factory = client._makeGetterFactory(
        url,
        client.HTTPClientFactory,
        contextFactory=contextFactory,
        *args, **kwargs)

    return factory.deferred.addErrback(handle206), factory

